@@ -1,5 +1,6 @@
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /* 
- *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
+ *  Copyright © 2002 Jorn Baayen <jorn@nl.linux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -13,23 +14,13 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- *  $Id: ephy-node-view.c,v 1.46.4.1 2004/11/07 19:57:02 chpe Exp $
  */
 
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif
 
-#include <gtk/gtktreeselection.h>
-#include <gtk/gtktreeviewcolumn.h>
-#include <gtk/gtkcellrenderertext.h>
-#include <gtk/gtkcellrendererpixbuf.h>
-#include <gtk/gtkcellrenderertoggle.h>
-#include <gtk/gtktreemodelfilter.h>
-#include <gtk/gtkwindow.h>
-#include <gtk/gtkmain.h>
+#include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
 
 #include "ephy-node-view.h"
@@ -38,8 +29,7 @@
 #include "ephy-dnd.h"
 #include "ephy-gui.h"
 #include "ephy-marshal.h"
-#include "ephy-string.h"
-#include "string.h"
+#include <string.h>
 
 static void ephy_node_view_class_init (EphyNodeViewClass *klass);
 static void ephy_node_view_init (EphyNodeView *view);
@@ -47,7 +37,7 @@ static void ephy_node_view_finalize (GObject *object);
 
 #define EPHY_NODE_VIEW_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), EPHY_TYPE_NODE_VIEW, EphyNodeViewPrivate))
 
-struct EphyNodeViewPrivate
+struct _EphyNodeViewPrivate
 {
 	EphyNode *root;
 
@@ -72,8 +62,6 @@ struct EphyNodeViewPrivate
 	gboolean remove_if_cancelled;
 	int editable_property;
 
-	int searchable_data_column;
-
 	gboolean drag_started;
 	int drag_button;
 	int drag_x;
@@ -84,6 +72,8 @@ struct EphyNodeViewPrivate
 	gboolean have_drag_data;
 	GtkSelectionData *drag_data;
 	guint scroll_id;
+
+	guint changing_selection : 1;
 };
 
 enum
@@ -92,6 +82,7 @@ enum
 	NODE_ACTIVATED,
 	NODE_SELECTED,
 	NODE_DROPPED,
+	NODE_MIDDLE_CLICKED,
 	LAST_SIGNAL
 };
 
@@ -111,11 +102,11 @@ static guint ephy_node_view_signals[LAST_SIGNAL] = { 0 };
 GType
 ephy_node_view_get_type (void)
 {
-	static GType ephy_node_view_type = 0;
+	static GType type = 0;
 
-	if (ephy_node_view_type == 0)
+	if (G_UNLIKELY (type == 0))
 	{
-		static const GTypeInfo our_info =
+		const GTypeInfo our_info =
 		{
 			sizeof (EphyNodeViewClass),
 			NULL,
@@ -128,12 +119,12 @@ ephy_node_view_get_type (void)
 			(GInstanceInitFunc) ephy_node_view_init
 		};
 
-		ephy_node_view_type = g_type_register_static (GTK_TYPE_TREE_VIEW,
-							      "EphyNodeView",
-							      &our_info, 0);
+		type = g_type_register_static (GTK_TYPE_TREE_VIEW,
+					       "EphyNodeView",
+					       &our_info, 0);
 	}
 
-	return ephy_node_view_type;
+	return type;
 }
 
 static void
@@ -162,16 +153,21 @@ static EphyNode *
 get_node_from_path (EphyNodeView *view, GtkTreePath *path)
 {
 	EphyNode *node;
-	GtkTreeIter iter, iter2;
+	GtkTreeIter iter, iter2, iter3;
 
 	if (path == NULL) return NULL;
 
 	gtk_tree_model_get_iter (view->priv->sortmodel, &iter, path);
 	gtk_tree_model_sort_convert_iter_to_child_iter
 		(GTK_TREE_MODEL_SORT (view->priv->sortmodel), &iter2, &iter);
+
+	if (iter2.stamp == 0) {
+		return NULL;
+	}
 	gtk_tree_model_filter_convert_iter_to_child_iter
-		(GTK_TREE_MODEL_FILTER (view->priv->filtermodel), &iter, &iter2);
-	node = ephy_tree_model_node_node_from_iter (view->priv->nodemodel, &iter);
+		(GTK_TREE_MODEL_FILTER (view->priv->filtermodel), &iter3, &iter2);
+
+	node = ephy_tree_model_node_node_from_iter (view->priv->nodemodel, &iter3);
 
 	return node;
 }
@@ -314,7 +310,8 @@ drag_motion_cb (GtkWidget *widget,
 				(node, view->priv->priority_prop_id);
 
 		if (priority != EPHY_NODE_VIEW_ALL_PRIORITY &&
-		    priority != EPHY_NODE_VIEW_SPECIAL_PRIORITY)
+		    priority != EPHY_NODE_VIEW_SPECIAL_PRIORITY &&
+		    ephy_node_get_is_drag_source (node))
 		{
 			action = context->suggested_action;
 		}
@@ -379,6 +376,10 @@ drag_data_received_cb (GtkWidget *widget,
 		return;
 	}	
 
+	/* appease GtkTreeView by preventing its drag_data_receive
+	* from being called */
+	g_signal_stop_emission_by_name (view, "drag_data_received");
+
 	if (!view->priv->have_drag_data)
 	{
 		view->priv->have_drag_data = TRUE;
@@ -389,27 +390,29 @@ drag_data_received_cb (GtkWidget *widget,
 	if (view->priv->drop_occurred)
 	{
 		EphyNode *node;
-		GList *uris;
+		char **uris;
 		gboolean success = FALSE;
 		GtkTreePath *path;
 
-		gtk_tree_view_get_dest_row_at_pos
-			(GTK_TREE_VIEW (widget), x, y, &path, &pos);
-
-		g_return_if_fail (path != NULL);
+		if (gtk_tree_view_get_dest_row_at_pos
+			(GTK_TREE_VIEW (widget), x, y, &path, &pos) == FALSE)
+		{
+			return;
+		}
 
 		node = get_node_from_path (view, path);
+		if (node == NULL) return;
 
-		uris = ephy_string_parse_uri_list (selection_data->data);
+		uris = gtk_selection_data_get_uris (selection_data);
 
-		if (uris != NULL)
+		if (uris != NULL && ephy_node_get_is_drag_dest (node))
 		{
 			/* FIXME fill success */
 			g_signal_emit (G_OBJECT (view),
 				       ephy_node_view_signals[NODE_DROPPED], 0,
 				       node, uris);
-			g_list_foreach (uris, (GFunc) g_free, NULL);
-			g_list_free (uris);
+			g_strfreev (uris);
+
 		}
 
 		view->priv->drop_occurred = FALSE;
@@ -421,12 +424,6 @@ drag_data_received_cb (GtkWidget *widget,
 			gtk_tree_path_free (path);
 		}
 	}
-
-
-	/* appease GtkTreeView by preventing its drag_data_receive
-	 * from being called */
-	g_signal_stop_emission_by_name (GTK_TREE_VIEW (view),
-					"drag_data_received");
 }
 
 static gboolean
@@ -448,7 +445,7 @@ drag_drop_cb (GtkWidget *widget,
 
 void
 ephy_node_view_enable_drag_dest (EphyNodeView *view,
-				 GtkTargetEntry *types,
+				 const GtkTargetEntry *types,
 				 int n_types)
 {
 	GtkWidget *treeview;
@@ -507,9 +504,13 @@ static void
 ephy_node_view_selection_changed_cb (GtkTreeSelection *selection,
 			             EphyNodeView *view)
 {
+	EphyNodeViewPrivate *priv = view->priv;
 	GList *list;
 	EphyNode *node = NULL;
 
+	/* Work around bug #346662 */
+	if (priv->changing_selection) return;
+
 	list = ephy_node_view_get_selection (view);
 	if (list)
 	{
@@ -565,18 +566,35 @@ path_toggled (GtkTreeModel *dummy_model, GtkTreePath *path,
 		       node, checked);
 }
 
+static EphyNode *
+process_middle_click (GtkTreePath *path,
+		      EphyNodeView *view)
+{
+	EphyNode *node;
+	GtkTreeIter iter, iter2;
+	
+	gtk_tree_model_get_iter (view->priv->sortmodel, &iter, path);
+	gtk_tree_model_sort_convert_iter_to_child_iter
+		(GTK_TREE_MODEL_SORT (view->priv->sortmodel), &iter2, &iter);
+	gtk_tree_model_filter_convert_iter_to_child_iter
+		(GTK_TREE_MODEL_FILTER (view->priv->filtermodel), &iter, &iter2);
+
+	node = ephy_tree_model_node_node_from_iter (view->priv->nodemodel, &iter);
+	
+	return node;
+}
+
 static gboolean
 ephy_node_view_key_press_cb (GtkTreeView *treeview,
 			     GdkEventKey *event,
 			     EphyNodeView *view)
 {
-	guint32 unicode;
-
-	unicode = gdk_keyval_to_unicode (event->keyval);
+	gboolean handled = FALSE;
 
 	if (event->keyval == GDK_space ||
 	    event->keyval == GDK_Return ||
-	    event->keyval == GDK_KP_Enter)
+	    event->keyval == GDK_KP_Enter ||
+	    event->keyval == GDK_ISO_Enter)
 	{
 		if (view->priv->toggle_column >= 0)
 		{
@@ -585,15 +603,11 @@ ephy_node_view_key_press_cb (GtkTreeView *treeview,
 			selection = gtk_tree_view_get_selection (treeview);
 			gtk_tree_selection_selected_foreach
 					(selection, path_toggled, view);
+			handled = TRUE;
 		}
 	}
-	else if (view->priv->searchable_data_column != -1 && unicode)
-	{
-		return ephy_gui_select_row_by_key
-			(treeview, view->priv->searchable_data_column, unicode);
-	}
 
-	return FALSE;
+	return handled;
 }
 
 static void
@@ -673,6 +687,42 @@ did_not_drag (EphyNodeView *view,
         }                                                                                                              
 }
 
+typedef struct
+{
+	EphyNodeView *view;
+	gboolean result;
+}
+ForeachData;
+
+static void
+check_node_is_drag_source (GtkTreeModel *model,
+			   GtkTreePath *path,
+			   GtkTreeIter *iter,
+			   ForeachData *data)
+{
+	EphyNode *node;
+
+	node = get_node_from_path (data->view, path);
+	data->result = data->result &&
+		       node != NULL &&
+		       ephy_node_get_is_drag_source (node);
+}
+
+static gboolean
+can_drag_selection (EphyNodeView *view)
+{
+	GtkTreeView *tree_view = GTK_TREE_VIEW (view);
+	GtkTreeSelection *selection;
+	ForeachData data = { view, TRUE };
+
+	selection = gtk_tree_view_get_selection (tree_view);
+	gtk_tree_selection_selected_foreach (selection,
+					     (GtkTreeSelectionForeachFunc) check_node_is_drag_source,
+					     &data);
+
+	return data.result;
+}
+
 static void
 drag_data_get_cb (GtkWidget *widget,
                   GdkDragContext *context,
@@ -717,6 +767,7 @@ button_release_cb (GtkWidget *widget,
                         did_not_drag (view, event);
                         return TRUE;
                 }
+		view->priv->drag_started = FALSE;
         }
         return FALSE;
 }
@@ -737,11 +788,12 @@ motion_notify_cb (GtkWidget *widget,
 	{
 		if (gtk_drag_check_threshold (widget, view->priv->drag_x,
                                               view->priv->drag_y, event->x,
-					      event->y))
+					      event->y)
+		    && can_drag_selection (view))
 		{
 			context = gtk_drag_begin
 				(widget, view->priv->source_target_list,
-                                 GDK_ACTION_COPY,
+                                 GDK_ACTION_ASK | GDK_ACTION_COPY | GDK_ACTION_LINK,
                                  view->priv->drag_button,
                                  (GdkEvent*)event);
 
@@ -770,6 +822,11 @@ ephy_node_view_button_press_cb (GtkWidget *treeview,
 	GtkTreeSelection *selection;
 	gboolean call_parent = TRUE, path_is_selected;
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
+	
+	if (event->window != gtk_tree_view_get_bin_window (GTK_TREE_VIEW (treeview)))
+	{
+		return GTK_WIDGET_CLASS (parent_class)->button_press_event (treeview, event);
+	}
 
 	if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (treeview),
 					   event->x,
@@ -807,6 +864,14 @@ ephy_node_view_button_press_cb (GtkWidget *treeview,
 
 			g_signal_emit_by_name (view, "popup_menu", &retval);
 		}
+		else if (event->button == 2)
+		{
+			EphyNode *clicked_node;
+			
+			clicked_node = process_middle_click (path, view);
+			g_signal_emit (G_OBJECT (view),
+				       ephy_node_view_signals[NODE_MIDDLE_CLICKED], 0, clicked_node);
+		}
 		else if (event->button == 1)
 		{
 			if (view->priv->toggle_column >= 0)
@@ -924,7 +989,6 @@ cell_renderer_edited (GtkCellRendererText *cell,
                       const char *new_text,
                       EphyNodeView *view)
 {
-	GValue value = { 0, };
 	GtkTreePath *path;
 	GtkTreeIter iter, iter2;
 	EphyNode *node;
@@ -943,12 +1007,8 @@ cell_renderer_edited (GtkCellRendererText *cell,
 		(GTK_TREE_MODEL_FILTER (view->priv->filtermodel), &iter, &iter2);
 	node = ephy_tree_model_node_node_from_iter (view->priv->nodemodel, &iter);
 
-	g_value_init (&value, G_TYPE_STRING);
-	g_value_set_string (&value, new_text);
-	ephy_node_set_property (node,
-			        view->priv->editable_property,
-			        &value);
-	g_value_unset (&value);
+	ephy_node_set_property_string (node, view->priv->editable_property,
+				       new_text);
 
 	gtk_tree_path_free (path);
 
@@ -1161,13 +1221,14 @@ ephy_node_view_add_data_column (EphyNodeView *view,
 	return column;
 }
 
-GtkTreeViewColumn *
+int
 ephy_node_view_add_column (EphyNodeView *view,
 			   const char  *title,
 			   GType value_type,
 			   guint prop_id,
 			   EphyNodeViewFlags flags,
-			   EphyTreeModelNodeValueFunc icon_func)
+			   EphyTreeModelNodeValueFunc icon_func,
+			   GtkTreeViewColumn **ret)
 
 {
 	GtkTreeViewColumn *gcolumn;
@@ -1207,12 +1268,6 @@ ephy_node_view_add_column (EphyNodeView *view,
 				  G_CALLBACK (renderer_editing_canceled_cb), view);
 	}
 
-	if ((flags & EPHY_NODE_VIEW_SEARCHABLE) &&
-	    (view->priv->searchable_data_column == -1))
-	{
-		view->priv->searchable_data_column = column;
-	}
-
 	gtk_tree_view_column_pack_start (gcolumn, renderer, TRUE);
 	gtk_tree_view_column_set_attributes (gcolumn, renderer,
 					     "text", column,
@@ -1236,10 +1291,33 @@ ephy_node_view_add_column (EphyNodeView *view,
 
 	if (flags & EPHY_NODE_VIEW_SORTABLE)
 	{
+		/* Now we have created a new column, re-create the
+		 * sort model, but ensure that the set_sort function
+		 * hasn't been called, see bug #320686 */
+		g_assert (view->priv->sort_column == -1);
+		g_object_unref (view->priv->sortmodel);
+		view->priv->sortmodel = ephy_tree_model_sort_new (view->priv->filtermodel);
+		gtk_tree_view_set_model (GTK_TREE_VIEW (view), GTK_TREE_MODEL (view->priv->sortmodel));
+
 		gtk_tree_view_column_set_sort_column_id (gcolumn, column);
 	}
 
-	return gcolumn;
+	if (flags & EPHY_NODE_VIEW_SEARCHABLE)
+	{
+		gtk_tree_view_set_search_column (GTK_TREE_VIEW (view), column);
+		gtk_tree_view_set_enable_search (GTK_TREE_VIEW (view), TRUE);
+	}
+	
+	if (flags & EPHY_NODE_VIEW_ELLIPSIZED)
+	{
+		g_object_set (renderer, "ellipsize-set", TRUE,
+			      "ellipsize", PANGO_ELLIPSIZE_END, NULL);
+	}
+
+	if (ret != NULL)
+		*ret = gcolumn;
+
+	return column;
 }
 
 void
@@ -1281,13 +1359,7 @@ ephy_node_view_init (EphyNodeView *view)
 {
 	view->priv = EPHY_NODE_VIEW_GET_PRIVATE (view);
 
-	view->priv->remove_if_cancelled = FALSE;
 	view->priv->toggle_column = -1;
-	view->priv->filter = NULL;
-	view->priv->editable_renderer = NULL;
-	view->priv->edited_node = NULL;
-	view->priv->searchable_data_column = -1;
-	view->priv->source_target_list = NULL;
 	view->priv->priority_column = -1;
 	view->priv->priority_prop_id = 0;
 	view->priv->sort_column = -1;
@@ -1333,11 +1405,13 @@ ephy_node_view_get_selection (EphyNodeView *view)
 void
 ephy_node_view_remove (EphyNodeView *view)
 {
+	EphyNodeViewPrivate *priv = view->priv;
 	GList *list, *l;
 	EphyNode *node;
-	GtkTreeIter iter, iter2;
+	GtkTreeIter iter, iter2, iter3;
 	GtkTreePath *path;
 	GtkTreeRowReference *row_ref = NULL;
+	GtkTreeSelection *selection;
 
 	/* Before removing we try to get a reference to the next node in the view. If that is
 	 * not available we try with the previous one, and if that is absent too,
@@ -1349,9 +1423,9 @@ ephy_node_view_remove (EphyNodeView *view)
 
 	node = g_list_first (list)->data;
 	ephy_tree_model_node_iter_from_node (EPHY_TREE_MODEL_NODE (view->priv->nodemodel),
-					     node, &iter);
+					     node, &iter3);
 	gtk_tree_model_filter_convert_child_iter_to_iter (GTK_TREE_MODEL_FILTER (view->priv->filtermodel),
-							  &iter2, &iter);
+							  &iter2, &iter3);
 	gtk_tree_model_sort_convert_child_iter_to_iter (GTK_TREE_MODEL_SORT (view->priv->sortmodel),
 							&iter, &iter2);
 	iter2 = iter;
@@ -1371,13 +1445,20 @@ ephy_node_view_remove (EphyNodeView *view)
 	}
 	gtk_tree_path_free (path);
 
+	/* Work around bug #346662 */
+	priv->changing_selection = TRUE;
 	for (l = list; l != NULL; l = l->next)
 	{
 		ephy_node_unref (l->data);
 	}
+	priv->changing_selection = FALSE;
 
 	g_list_free (list);
 
+	/* Fake a selection changed signal */
+	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
+	g_signal_emit_by_name (selection, "changed");
+
 	/* Select the "next" node */
 
 	if (row_ref != NULL)
@@ -1415,9 +1496,10 @@ ephy_node_view_select_node (EphyNodeView *view,
 
 void
 ephy_node_view_enable_drag_source (EphyNodeView *view,
-				   GtkTargetEntry *types,
+				   const GtkTargetEntry *types,
 				   int n_types,
-				   int column)
+				   int base_drag_column,
+				   int extra_drag_column)
 {
 	GtkWidget *treeview;
 
@@ -1428,8 +1510,10 @@ ephy_node_view_enable_drag_source (EphyNodeView *view,
 	view->priv->source_target_list =
 		gtk_target_list_new (types, n_types);
 
-	ephy_tree_model_sort_set_column_id (EPHY_TREE_MODEL_SORT (view->priv->sortmodel),
-					    column);
+	ephy_tree_model_sort_set_base_drag_column_id  (EPHY_TREE_MODEL_SORT (view->priv->sortmodel),
+					               base_drag_column);
+	ephy_tree_model_sort_set_extra_drag_column_id (EPHY_TREE_MODEL_SORT (view->priv->sortmodel),
+					               extra_drag_column);
 
 	g_signal_connect_object (G_OBJECT (view),
 			         "button_release_event",
@@ -1486,23 +1570,6 @@ ephy_node_view_is_target (EphyNodeView *view)
 	return gtk_widget_is_focus (GTK_WIDGET (view));
 }
 
-gboolean
-ephy_node_view_has_selection (EphyNodeView *view, gboolean *multiple)
-{
-	GtkTreeSelection *selection;
-	int rows;
-
-	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (view));
-	rows = gtk_tree_selection_count_selected_rows (selection);
-
-	if (multiple)
-	{
-		*multiple = rows > 1;
-	}
-
-	return rows > 0;
-}
-
 static gboolean
 filter_visible_func (GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
 {
@@ -1625,14 +1692,14 @@ ephy_node_view_class_init (EphyNodeViewClass *klass)
 					 g_param_spec_pointer ("root",
 							       "Root node",
 							       "Root node",
-							       G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+							       G_PARAM_READWRITE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_CONSTRUCT_ONLY));
 	g_object_class_install_property (object_class,
 					 PROP_FILTER,
 					 g_param_spec_object ("filter",
 							      "Filter object",
 							      "Filter object",
 							      EPHY_TYPE_NODE_FILTER,
-							      G_PARAM_READWRITE));
+							      G_PARAM_READWRITE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
 
 	ephy_node_view_signals[NODE_TOGGLED] =
 		g_signal_new ("node_toggled",
@@ -1676,6 +1743,16 @@ ephy_node_view_class_init (EphyNodeViewClass *klass)
 			      2,
 			      G_TYPE_POINTER,
 			      G_TYPE_POINTER);
+	ephy_node_view_signals[NODE_MIDDLE_CLICKED] =
+		g_signal_new ("node_middle_clicked",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (EphyNodeViewClass, node_middle_clicked),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__POINTER,
+			      G_TYPE_NONE,
+			      1,
+			      G_TYPE_POINTER);
 
 	g_type_class_add_private (object_class, sizeof (EphyNodeViewPrivate));
 }
